// Version 5.0 of October 2004
// Version 5.7 of October 2005
// All rights reserved
// Published under the GNU Public License

// Program originally written 1985 in HP-Basic by Ekkehart Schlicht
// Ported to C by Robert Schlicht in 2002.
// Various additions by Ekkehart Schlicht.
// Helpful suggestions by Johannes Ludsteck.

// To compile with the Microsoft C compiler, use a command 
// like this one:
// "C:\Program Files\Microsoft C\bin\cl" /Ox /I "C:\Program Files\Microsoft C\include"
//		   vcc.c /link /subsystem:windows /libpath:"C:\Progra~1\Micros~2\lib"
//		   kernel32.lib user32.lib gdi32.lib comdlg32.lib


// To compile under Linux use a command like
// gcc -o vcc -lm vcc.c
// and uncomment the following line:
// #define LINUX 1


#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <time.h>

// Some auxiliary functions for handling vectors and matrices.
// ===========================================================

// Die Funktion 'vector' erzeugt einen Vektor bzw. ndert seine Lnge.
// 'x' zeigt auf die Adresse des Vektors, die auf 0 gesetzt ist, wenn der
// Vektor die Lnge 0 hat, und die auf 0 gesetzt werden muss, bevor die
// Funktion das erste Mal fr einen Vektor aufgerufen wird. 'length' ist
// die neue Lnge des Vektors, die mitgespeichert wird. 'size' ist der
// Speicherbedarf eines einzelnen Elements in Bytes. Wenn nicht
// gengend Speicherplatz frei ist, wird das Programm abgebrochen.
// Das Makro 'LENGTH' ermittelt die Lnge eines Vektors:
	
#define LENGTH(x) ( !(x)? 0 : *((unsigned int*)(void*)(x) - 1) )

void vector(void ** x, unsigned int size, unsigned int length)
{
	unsigned int *ptr;
	ptr = !*x? 0 : (unsigned int*)*x - 1;			   // Anfang des Speicherbereichs ermitteln
	if (!length) {if (ptr) free(ptr); *x = 0; return;} // Speicherplatz freigeben, wenn length=0 
	if (ptr && *ptr == length) return;				   // Sonst prfen, ob Gre verndert wird
	ptr = realloc(ptr, sizeof *ptr + length * size);   // Gre des Speicherbereichs verndern
	if (!ptr) abort();								   // bei Fehler Programm abbrechen
	*x = (void*)(ptr + 1);							   // Vektoranfang auf zweites Element setzen
	*ptr = length;									   // im ersten Element Vektorlnge speichern
}


// Dieselbe Funktion fr eine Matrix (= Vektor aus Vektoren):

void matrix(void *** A, unsigned int size, unsigned int length, unsigned int width)
{
	unsigned int n;
	n = LENGTH(*A); 										 // alte Matrixlnge speichern
	while (n > length) vector(&(*A)[n], size, 0);			 // berflssige Zeilen lschen
	vector((void**)A, sizeof **A, length);					 // Matrixlnge anpassen
	while (n < length) (*A)[n++] = 0;						 // zustzliche Zeilen initialisieren
	for (n=0; n<length; n++) vector(&(*A)[n], size, width);  // Zeilenlngen anpassen
}


// Die folgenden beiden Makros erleichtern die Verwendung von vector und matrix:

#define VECTOR(x, length) vector((void**)&(x), sizeof *(x), (length))
#define MATRIX(A, length, width) matrix((void***)&(A), sizeof **(A), (length), (width))


// Definition of program options taken from the VC-Program (Windows version)
// =========================================================================

// Programm-Optionen (globale Variablen)
// check1 controls adjustment
// check3 flag for calculating covariance matrix
// check4 take initial variances from input
// check5 no intercept
// check6 take random initial values for variances
// check7 use likelihoosd estimator instead of moments estimator
// radio2 estimate selected variances only
// radio3 don't estimate variances
// *check indicates which variances to freeze (not adjust)
// converged flag indicating convergence

int check1, check3, check4, check5, check6, check7, *check, radio2, radio3, convergence;
static double **C;

// Main program
// ============

int main(int argc, char *argv[])
{
	void sol(double **x, double *y, double **A, double *r, double *z);
	void standarderror(double **C, double *as);
	void errormatrix(double **c, double *as);
	double var(double **x, double *y, double **A, double *r, double *s0, double *s, double *z);

	FILE *file;
	double **x=0, *y=0, **A=0, *as=0, **AS=0,*r=0, *z=0, *s=0, *sini=0, s0, s0ini;
	double v, precision;

	unsigned int l, n, n5, i, j, t, arguments, understood, zerovar, negvar,  numeric, maxit;
	char c, name[1024], infile[1024], outfile[1024], helptext[1024], message[1024],  **rname=0, **cname = 0;
	char * pEnd;
// Help text
     strcpy(helptext, "\nvcc - a program for estimating time-varying coefficients\n");
		 strcat(helptext,"by Ekkehart Schlicht, version 5.7 \nPublished under the GNU Public License.\n");
		 strcat(helptext,"All rights reserved. Available at www.lrz.de/~ekkehart\n\n");
         strcat(helptext,"Usage: vcc inputfile [switches]\n\n");
		 strcat(helptext," inputfile    specifies the input file (CSV format - see documentation)\n");
		 strcat(helptext,"      --h     This help screen.\n");
		 strcat(helptext,"      --ni    Don't add an intercept term\n");
		 strcat(helptext,"      --r     take random initial values for variances\n");
	 	 strcat(helptext,"      --v     Take initial variances from input file.\n");
		 strcat(helptext,"              Negative variance inputs will be frozen to their absolute values.\n");
		 strcat(helptext,"              Zero variance inputs will be frozen close to zero. Requires --ni\n");
         strcat(helptext,"      --ml    Use maximum likelihood estimator instead of moments estimator.\n");
		 strcat(helptext,"      --f     Don't estimate variances, just filter. Requires --ni\n");
		 strcat(helptext,"      --cv    calculate the complete covariance matrix\n");
		 strcat(helptext,"      --s     Slow down adjustment speed to improve numerical stability.\n");
		 strcat(helptext,"      --p n   Set precision n. Default is 5 which means 10^-5\n");
		 strcat(helptext,"      --it n  Set maximum number of iterations n. Default is 100000\n\n");



// Evaluate command line
	
		 arguments=argc;
		 numeric=0;
		 precision=0.00001;
		 maxit=100000;

	if (argc==1) { printf(helptext); return 0;};
	if (argc==2) { if((strcmp(argv[1],"--h")==0)) printf(helptext); else strcpy(infile,argv[1]);};
	if (argc > 2) { 
	       for(i=2; i<arguments;i++) {
			   if(numeric){
				   numeric=0;
			   understood=1;
			   }
			   else {
				   understood=0;
			   if (strcmp(argv[i],"--h")==0){printf(helptext); understood=1;};
			   if (strcmp(argv[i],"--s")==0){understood=1; };
			   if (strcmp(argv[i],"--v")==0){understood=1;check4=1;};
			   if (strcmp(argv[i],"--f")==0){understood=1;radio3=1;};
               if (strcmp(argv[i],"--ni")==0){understood=1;check5=1;};
			   if (strcmp(argv[i],"--r")==0){understood=1;check6=1;};
			   if (strcmp(argv[i],"--ml")==0){understood=1;check7=1;};
               if (strcmp(argv[i],"--cv")==0){understood=1;check3=1; };
			   if (strcmp(argv[i],"--p")==0){
				   precision=atof(argv[i+1]);
				   precision=pow(10.,(-precision));
				   numeric=1;
				   understood=1; };
			  
				if (strcmp(argv[i],"--it")==0){
				   maxit = strtoul(argv[i+1],&pEnd,10);
				   numeric=1;
				   understood=1; };

				   if (!understood) {printf("\nError: Cannot understand all command line switches.") ;
								printf("\nFor help, type vcc\n");
				  
				   return 1;}
			   }
			
		   };


	}
			
;
	

	if ((radio2||radio3||check4)&&(!check5)) {
				strcpy(message,"\n          If initial variances are to be taken from input, \n");
				strcat(message,"          the constant for the intercept must be provided \n");
				strcat(message,"          in the input file and the switch --ni (no intercept)\n");
				strcat(message,"          must be checked as well. Otherwise I will not know which\n");
				strcat(message,"          variance belongs to which coefficient.\n\n");
				strcat(message,"          I shut down now.\n\n");
				strcat(message,"          Check that your input file provides a constant in its\n");
				strcat(message,"          third or later column (unless you don't want one),\n");
				strcat(message,"          add the switch --ni and try again.\n\n");
		printf(message); return 0;
	}
	
	zerovar=0;								// flag for zero variances in input
	negvar=0;								// flag for negative variance input
	
	
// Open input file

	    strcpy(infile,argv[1]);
		file = fopen(infile, "rt");
    if (!file) {
		printf("\nFile opening error\n");
		return 1;  
		}									// File opening error
    strcpy(outfile,infile);
	outfile[strlen(outfile)-4]='\0';		// truncate filename



	strcat(outfile,"-VC.csv");				// add extension for output file name


// Read data from file. (Matrix x und Vektor y erzeugen, Zeilennamen rname einlesen)
	
	n = 0;
	do {													// in erster Zeile Spalten zhlen
		fscanf(file, "%[^,\n]%c ", name, &c);				// und Spaltennamen einlesen
#ifdef LINUX
		if (c == '\n') name[strlen(name)-1]='\0';			// remove line feed for LINUX
#endif
		VECTOR(cname, n+1);
		cname[n] = 0; VECTOR(cname[n], strlen(name) + 1);
		strcpy(cname[n], name);
		++n;		
	} while (c == ',');
	if (n < 3) return 1;
	
		if (!check5) {										// mit Variable "intercept" ergnzen
		VECTOR(cname, n+1);
		cname[n] = 0; VECTOR(cname[n], strlen(name) + 1);
		cname[n]="intercept";
		
		}
	
		n -= 2;	 								// n = Breite von x (erste 
                                                // Spalte wird ignoriert, zweite Spalte ist y)	
	 

	if (!check5) n5=n+1; else n5=n;				// we need one column more for the constant

	VECTOR(s, n5);								// Varianzvektor s erzeugen
    VECTOR(sini,n5);							// Anfangswerte der Varianzen
	VECTOR(check,n5);							// Flagvektor erzeugen und auf Nullsetzen
	for (i=0; i<n; i++) {check[i]=0;};
		
	if ((!check4)||(!check6)) {
	   	s0 = 10;
		s0ini=s0;								// Anfangswert speichern
	
			for (i=0; i<n5; i++) {
				s[i]=1;							// Anfangswerte fr Varianzen s einlesens einlesen
				sini[i]=s[i];					// Anfangswert speichern
			}
		};
	if (check6) {                               // Randomize initial values
		srand( (unsigned)time( NULL ) );
		s0=rand();
		s0ini=s0;
			for (i=0; i<n5; i++) {
				s[i]=rand();							// Anfangswerte fr Varianzen s einlesens einlesen
				sini[i]=s[i];					// Anfangswert speichern
			}
		}
	if (check4) {                               // Take initial variances from input
		fscanf(file, "%*[^,],%lg ", &s0);		// s0 einlesen
		if (s0 < 0) {
			printf("\nVariance of the disturbance in the equation cannot be frozen.");
			printf("\nI take its absolute value.\n");
			s0 = -s0; 
		}
		if(s0==0) {
			s0=10.;
		};
			s0ini=s0;							// Anfangswert speichern

		for (i=0; i<n; i++) {
           
					fscanf(file, ",%lg ", &s[i]);       // Anfangswerte fr Varianzen s einlesens einlesen
					if (s[i]==0) {
						s[i] = 1e-10;					// if zero variance, make it positive and freeze it
					    check[i]=1;
						radio2=1;
                        zerovar=1;
					}	
					if (s[i]<0) {
						s[i] = -s[i];					// if negative variance, make it positive and freeze it
					    check[i]=1;
						radio2=1;
                        negvar=1;
					}							
				sini[i]=s[i];							// Save initial variances
				
				}
	}


// Messages on variance input

	if (zerovar){
		strcpy(message,"\nSome initial variances are zero. I have frozen");
		strcat(message,"\nthem close to zero to enforce constancy of the ");
		strcat(message,"\nassociated coefficient.\n");
		printf(message);
	}
	
	
	if (negvar){
		strcpy(message, "\nSome initial variances were negative. I have frozen");
		strcat(message,"\nthem to their absolute values.\n");
		printf(message);
	}		
	
	
	
		sprintf(name, "Initial values of variances: %lg", s0);
			for (i=0; i<n; i++) sprintf(name + strlen(name), ", %lg", s[i]);
	
				
		l = 0;													// l = Lnge von x bzw. y

		
		while (!feof(file)) {										
			VECTOR(y, l+1), MATRIX(x, l+1, n5);					// Lnge von x und y vergrern
			VECTOR(rname, l+1); 								// Lnge von rname vergrern
			fscanf(file, "%[^,],%lg ", name, &y[l]);			// erste Spalte (Zeilennamen) und y-Wert einlesen
			rname[l] = 0; VECTOR(rname[l], strlen(name) + 1);	// Zeilennamen abspeichern
			strcpy(rname[l], name);
			for (i=0; i<n; i++) fscanf(file, ",%lg ", &x[l][i]);// Zeile von x einlesen
			l++;												// l um 1 erhhen
		}

	
		if (!check5) {
			for (i=0; i<l; i++)
				x[i][n5-1]=1.;
		n=n+1;
		}
	// Close input file
	fclose(file);										

	// Vektoren r, z und Matrix A erzeugen
	i=l*n;

	// (Der "Varianzvektor (s(0), s(1), ..., s(N))" ist hier s0, s[0], ..., s[n-1].)
	VECTOR(r, n), VECTOR(z, n); MATRIX(A, l, n); MATRIX(AS, l, n); VECTOR(as, i);
	
	// Anfangswerte fr r festlegen
	for (i=0; i<n; i++) r[i] = s[i] / s0;

	
	// Normalgleichungssystem fr A lsen und iterierte
	// Varianzen und Varianzverhltnisse bestimmen

	// This is the main program loop
	// =============================
		
	i = 0;

	do {
	sol(x, y, A, r, z);														// solve normal equation, retur address of C


		for (j=0; j<n; j++) if 	(radio2&&check[j]) s[j]=sini[j];			// force selected variances to given values
		
		v = var(x, y, A, r, &s0, s, z);
	    


//		if (i==1) {
//			printf("\nIteration : %u,   error: %-.6lG  per cent.", i, v*100);
//		
//			
//		}
//		if (!(++i%100)) {
			i=i+1;
			printf("\rIteration : %u,   error: %-.6lf  per cent.                       ", i, v*100);
		
//		}
	} while ((v >= precision) &&(i<maxit)&&(!radio3)) ;

	// end main program loop

	if(i==maxit) {
		convergence=0;
		printf("\nRequired precision of %g per cent not reached after %u iterations.", precision*100, maxit);
		printf("\nIncrease number of iterations with the --it switch,");
		printf("\nIf results do not converge, try the --s switch.\n");
		}
	else{
		convergence=1;
		printf("\rAccuracy of %-.6lG  per cent. obtained after %u iterations.\n", v*100,i);
	}

	
	standarderror(C, as);
	for (i=0;i<n*l;i++) as[i]=sqrt(s0*as[i]);


	// Ergebnis in Datei speichern und Programm beenden
	// Write input and output to output file
	
	file = fopen(outfile, "wt");
	if (!file){ printf("\nError opening output file\n"); return 1;}			// (Fehler beim Erzeugen der Datei)

	fprintf(file, "%s,%s", cname[0], cname[1]);
	for (i=0; i<n; i++) fprintf(file, ",%s", cname[i+2]);
	for (i=0; i<n; i++) fprintf(file, ",a_%s,stderr_%s,lb_%s,ub_%s", cname[i+2],cname[i+2],cname[i+2],cname[i+2]);
	fprintf(file, "\n");
	
	for (t=0; t<l; t++) {
		fprintf(file, "%s,%lg", rname[t], y[t]);
		for (i=0; i<n; i++) fprintf(file, ",%lg", x[t][i]);
		for (i=0; i<n; i++) fprintf(file, ",%lg,%lg,%lg,%lg", A[t][i],as[t*n+i],A[t][i]-2*as[t*n+i],A[t][i]+2*as[t*n+i]);
		fprintf(file, "\n");
	}
	fprintf(file, "\n");
	fprintf(file, "\n");
	fprintf(file, "\n");
//	fprintf(file, "Initial variances");
//	fprintf(file,"\n");

	fprintf(file, "Variances,,equation");
	for (i=0; i<n; i++) fprintf(file, ",%s", cname[i+2]);

	fprintf(file, "\n\nInitial variances,,%lg", s0ini);
	for (i=0; i<n; i++) fprintf(file, ",%lg", sini[i]);

	fprintf(file, "\nCalculated variances");
	fprintf(file, ",,%lg", s0);
	for (i=0; i<n; i++) fprintf(file, ",%lg", s[i]);
	fprintf(file, "\n");
	if (radio3) fprintf(file, "\nNo iteration. Time series just filtered with initial variances.");   
	if (radio2){
		fprintf(file,"\n");	
		for (i=0; i<n; i++) {
				if 	(check[i]) {
				fprintf(file, "\nVariance of coefficient for %s frozen to initial value", cname[i+2]);
				}
			}
		}
	if (!radio3) fprintf(file, "\nRelative error in iterated variances:  %G per cent.", v*100);
    
	if (!convergence){				// Warning for lack of convergence
		fprintf(file,"\n\nWARNING! \nRequired precision of %g per cent not reached after %u iterations.", precision*100, maxit);
		fprintf(file,"\nIncrease number of iterations with the --it switch,");
		fprintf(file,"\nIf results do not converge try the --s switch.\n");
			}
    if (check7) fprintf(file, "\n\nEstimation method: Likelihood estimator"); 
	    else fprintf(file, "\n\nEstimation method: Moments estimator"); 
	
	fclose(file);
	printf("\nOutput saved to ");
	printf(outfile);
	printf("\n");


	// Write covariance matrix

	if (check3) {
	
		unsigned int i, j, n, m, k; 
		n=LENGTH(C);								// Zeilen
		m=LENGTH(*C);								// Spalten	
 		k=(n-1)*(m-1);

		outfile[strlen(outfile)-4]='\0';			// truncate filename
		strcat(outfile,"-covariances.csv");         // add extension for output file name
			
			file = fopen(outfile, "wt");
			if (!file) {
				printf("\nCould not generate output file for the covariance matrix.\n");
				return 1;	}						// File creation error
			
			for (i=0;i<n ;i++){
				for (j=0;j<n ;j++) as[j]=0;
				as[i]=1;
				errormatrix(C,as);
				for (j=0;j<(n-1) ;j++) fprintf(file, "%lg,",s0*as[j]);
				fprintf(file, "%lg\n",s0*as[n-1]);
			}

			fclose(file);
			printf("\nCovariance matrix saved to ");
			printf(outfile);
			printf("\n");
	}	
	
	return 0;
}

// Finished


// Routines
// ========

// Programm zur Lsung der Normalgleichungen.
// x, y und r werden vorgegeben, A wird bestimmt.

void sol(double **x, double *y, double **A, double *r, double *z)
{
	void cholband(double **A);
	void trace(double **C, double *z);
	void bandsol(double **A, double *x);

//	static double **C, *a;	// (a und C werden nur einmal beim ersten Funktionsaufruf erzeugt.)
		static double *a;	
		unsigned int l, n, t, i, j, q;
	double p;

	// Feldgrenzen von A sind l und n;
	l = LENGTH(x);
	n = LENGTH(*x);
	
	
	if (!C) MATRIX(C, l*n, n+1);

	for (j=1; j<l*n; j++) for (i=0; i<n+1; i++) C[j][i] = 0;

	for (t=0; t<l; t++) {
		p = y[t];
		q = t * n;
		for (i=0; i<n; i++) {
			A[t][i] = x[t][i] * p;
			for (j=0; j<=i; j++) C[q+i][n-i+j] = x[t][i] * x[t][j];
		}
	}
	
	for (i=0; i<n; i++) r[i] = 1/r[i];
	
	for (t=1; t<l; t++) {
		q = t * n;
		for (i=0; i<n; i++) C[q+i][0] = -r[i], C[q+i][n] += 2*r[i];
	}
	
	for (i=0; i<n; i++) C[i][n] += r[i], C[q+i][n] -= r[i], r[i] = 1/r[i];
	
	cholband(C);
	trace(C, z);

	// Matrix A in Vektor a umdimensionieren
	if (!a) VECTOR(a, l*n);
	{double *g; g = a; for (t=0; t<l; t++) for (i=0; i<n; i++) *g++ = A[t][i];}

	bandsol(C, a);
	
	// Umdimensionieren rckgngig machen
	{double *g; g = a; for (t=0; t<l; t++) for (i=0; i<n; i++) A[t][i] = *g++;}

}




// Programm zur Bestimmung der Standardabweichungen 
void standarderror(double **C, double *as)
{
	void bandsol(double **A, double *as);
	void bandtrn(double **A);
	void bandtrninv(double**A);
	unsigned int i, j, n, m; 
	double *e=0;

	n=LENGTH(C);													// Zeilen
																	
	m=LENGTH(*C);													// Spalten
	
	
	VECTOR(e,n);

	
		for (i=0; i<n; i++) { 
			for (j=0; j<n; j++) e[j]=0; 
				e[i]=1;
				bandtrninv(C);	
				bandsol(C,e);
				as[i]=e[i];
		}


}

// Programm zur Bestimmung einer Zeile der Kovarianzmatrix 
void errormatrix(double **C, double *as)
{
	void bandsol(double **A, double *as);
	void bandtrn(double **A);
	void bandtrninv(double**A);
		
		bandtrninv(C);	
		bandsol(C,as);
		
}

// A wird mit der Transponierten von A (kurz gespeichert)
// berschrieben
void bandtrn(double **A)
{
	unsigned int m, n, i, j;

	n = LENGTH(A);
	m = LENGTH(*A);
	
	for (i=0; i<n; i++) for (j=0; j<m && j<n-i; j++) A[i][j] = A[i+j][m-1-j];
	for (j=1; j<m; j++) for (i=n-j; i<n; i++) A[i][j]=0;
}

// Umkehrtng von bandtrn, A wird mit der Transponierten von A (kurz gespeichert)
// berschrieben
void bandtrninv(double **A)


{
	unsigned int m, n, i, j;

	n = LENGTH(A);
	m = LENGTH(*A);
				
	for (i=0; i<n; i++) for (j=0; j<m && j<n-i; j++) A[n-1-i][m-1-j] = A[n-1-i-j][j];
	for (j=1; j<m; j++) for (i=n-j; i<n; i++) A[n-1-i][m-1-j]=0;
}

// bandsol1 lst A*y=x und berschreibt x mit y:
// A ist untere Dreiecksmatrix kurz gespeichert
void bandsol1(double **A, double *x)
{
	unsigned int m, n, p, i, j;
	n = LENGTH(A);
	m = LENGTH(*A);
	p = m-1;
	
	for (i=0; i<n; i++) {
		for (j = i<p? p-i:0; j<p; j++) x[i] -= A[i][j] * x[j-(p-i)];
		x[i] /= A[i][p];
	}


}

// Wie bandsol1, nur fr obere Dreiecksmatrix
void bandsol2(double **A, double *x)
{
	unsigned int m, n, i, j;

	n = LENGTH(A);
	m = LENGTH(*A);
	
	for (i=n; i>0;) {i--;
		for (j=1; j<m && j<n-i; j++) x[i] -= A[i][j] * x[i+j];
		x[i] /= A[i][0];
	}
}

// bandsol lst (A*A')*y=x und berschreibt x mit y.
// A ist untere ist untere Dreiecksmatrix, kurz gespeichert.
void bandsol(double **A, double *x)
{
	bandsol1(A, x);
	bandtrn(A);
	bandsol2(A, x);

}

// Cholesky-Zerlegung der (kurz gespeicherten) Bandmatrix A
// A wird mit unterer Dreiecksmatrix berschrieben.
void cholband(double **A)
{
	unsigned int n, m, p, i, j, k;
	
	n = LENGTH(A);
	m = LENGTH(*A);
	p = m-1;

	for (i=0; i<n; i++) {
		for (j = i>p? i-p:0; j<i; j++) {
			for (k = i>p? i-p:0; k<j; k++) A[i][p+j-i] -= A[i][p+k-i]*A[j][p+k-j];
			A[i][p+j-i] /= A[j][p];
		}
		for (k = i>p? i-p:0; k<i; k++) A[i][p] -= A[i][p+k-i] * A[i][p+k-i];
		A[i][p] = sqrt(A[i][p]);
	}
}

// trace bestimmt z(i) = tr{P(i)*INV(CC')*P'(i)}
// fr i = 1 bis n, C untere Dreiecksmatrix, kurz gespeichert.
void trace(double **C, double *z)
{
	unsigned int m, n, l, i, j, t;
	static double *x, *y; // (Diese Vektoren werden nur einmal beim ersten Funktionsaufruf erzeugt.)
	double q;
	
	m = LENGTH(C);
	n = LENGTH(z);
	l = m / n;
	
	if (!x) VECTOR(x, m);
	if (!y) VECTOR(y, m);
	
	for (i=0; i<n; i++) {
		q = 0;
		for (j=0; j<m; j++) y[j] = 0;
		y[i] = 1;
		bandsol1(C, y);
		for (t=1; t<l; t++) {
			for (j=0; j<m; j++) x[j] = y[j], y[j] = 0;
			y[t*n+i] = 1;
			bandsol1(C, y);
			for (j=0; j<m; j++) x[j] = y[j] - x[j], q += x[j] * x[j];
		}
		z[i] = q;
	}

}

// var bestimmt neue Varianzverhltnisse und Varianzen
double var(double **x, double *y, double **A, double *r, double *s0, double *s, double *z)
{
	static double *u, *r0; // (Diese Vektoren werden nur einmal beim ersten Funktionsaufruf erzeugt.)
	unsigned int l, n, t, i, con;
	double p, q, f;
	
	l = LENGTH(x);
	n = LENGTH(*x);
	if (check7) con=l-1;  else  con = l-n;         // for likelihood estimator or moments estimator
	if (!u) VECTOR(u, l);
	if (!r0) VECTOR(r0, n);

	for (i=0; i<n; i++) r0[i] = r[i];

	for (t=0; t<l; t++) {
		u[t] = -y[t];
		for (i=0; i<n; i++) u[t] += A[t][i]*x[t][i];
	}

	q = 0; for (t=0; t<l; t++) q += u[t] * u[t];

	for (i=0; i<n; i++) {
		s[i] = 0;
		for (t=1; t<l; t++) f = A[t][i] - A[t-1][i], s[i] += f * f;
		q += s[i] / r[i];
	}
	*s0 = q / (l-n);

	p = 0;
	for (i=0; i<n; i++) {
		r[i] = (s[i] * con / q + z[i]) / (l-1);								// calculate new variance ratios
		if 	(!(radio2&&check[i])) {												// skip updating blocked variance ratios
				f = r[i] / r0[i];

		if (!check1) r[i] = r0[i] * pow(f, 20);									// fast relaxation, disabled by check1
		s[i] = r[i] * *s0;
		f -= 1; if (f > p) p = f; else if (-f > p) p = -f;
	}

	}
	return p;
}

